package main

import (
	"os/exec"
	"strings"

	"github.com/go-git/go-git/v5"
	"github.com/go-git/go-git/v5/plumbing/object"
	"github.com/go-git/go-git/v5/plumbing/storer"
	"github.com/pkg/errors"
)

/*n, err := LoadNotebook("/home/trevor/stick/personal/")
if err != nil {
	panic(err)
}

log.Print(time.Unix(n.Created(), 0))
log.Print(time.Unix(n.Modified(), 0))

err = n.Upload()
if err != nil {
	if strings.Contains(err.Error(), git.ErrNonFastForwardUpdate.Error()) {
		log.Println("conflict while pushing")
	} else if err != git.NoErrAlreadyUpToDate {
		panic(err)
	}
}
err = n.Download()
if err != nil {
	if strings.Contains(err.Error(), git.ErrNonFastForwardUpdate.Error()) {
		log.Println("conflict while updating")
		log.Println(err)
	} else if err != git.NoErrAlreadyUpToDate {
		panic(err)
	}
}*/

const DefaultSubmoduleDepth = 10

type Notebook struct {
	ID    string
	Label string
	Serve map[string]int `json:"-"`

	Repository *git.Repository `json:"-"`
}

func (n *Notebook) getNote(id string) *Note {
	file := n.File(id)
	if file == nil {
		return nil
	}

	r := n.Repository

	ref, err := r.Head()
	if err != nil {
		return nil // Empty repository
	}

	cIter, err := r.Log(&git.LogOptions{From: ref.Hash(), Order: git.LogOrderCommitterTime, FileName: &file.Name})
	CheckError(err)

	note := &Note{ID: id}

	err = cIter.ForEach(func(c *object.Commit) error {
		note.ModifiedAt = c.Author.When.Unix()
		note.ModifiedBy = c.Author.Name

		return storer.ErrStop
	})
	CheckError(err)

	note.Label = file.Name
	if strings.HasSuffix(strings.ToLower(note.Label), ".md") {
		note.Label = note.Label[0 : len(note.Label)-3]
	}
	note.Label = strings.Title(strings.Replace(note.Label, "-", " ", -1))

	note.Body, err = file.Contents()
	if err != nil {
		return nil
	}

	return note
}

func (n *Notebook) allNotes() map[string]*Note {
	r := n.Repository

	ref, err := r.Head()
	CheckError(err)

	commit, err := r.CommitObject(ref.Hash())
	CheckError(err)

	tree, err := commit.Tree()
	CheckError(err)

	notes := make(map[string]*Note)
	err = tree.Files().ForEach(func(f *object.File) error {
		noteID := hash(f.Name)
		note := n.getNote(noteID)
		notes[noteID] = note
		return nil
	})
	CheckError(err)

	return notes
}

type ServedNotebook struct {
	*Notebook
	Notes map[string]*Note
}

func NewNotebook(worktree string, cloneurl string) (*Notebook, error) {
	var err error

	n := &Notebook{}
	n.Repository, err = InitializeRepository(worktree, cloneurl)
	if err != nil {
		return nil, errors.Wrap(err, "failed to initialize repository")
	}

	return n, nil
}

func LoadNotebook(worktree string) (*Notebook, error) {
	var err error

	n := &Notebook{}
	n.Repository, err = LoadRepository(worktree)
	if err != nil {
		return nil, errors.Wrap(err, "failed to load repository")
	}

	return n, nil
}

func (n *Notebook) Created() int64 {
	head, err := n.Repository.Head()
	if err != nil {
		return 0 // No commits yet
	}

	commit, err := n.Repository.CommitObject(head.Hash())
	if err != nil {
		return 0
	}

	for {
		numParents := commit.NumParents()
		if numParents == 0 {
			break
		}

		commit, err = commit.Parent(numParents - 1)
		if err != nil {
			return 0
		}
	}

	return commit.Author.When.Unix()
}

func (n *Notebook) Download() error {
	w, err := n.Repository.Worktree()
	if err != nil {
		return errors.Wrap(err, "failed to read worktree")
	}

	return w.Pull(&git.PullOptions{RemoteName: "origin", RecurseSubmodules: DefaultSubmoduleDepth})
}

func (n *Notebook) Upload() error {
	return n.Repository.Push(&git.PushOptions{RemoteName: "origin"})
}

func (n *Notebook) File(id string) *object.File {
	var file *object.File

	ref, err := n.Repository.Head()
	if err != nil {
		return nil // Empty repository
	}

	commit, err := n.Repository.CommitObject(ref.Hash())
	CheckError(err)

	tree, err := commit.Tree()
	CheckError(err)

	err = tree.Files().ForEach(func(f *object.File) error {
		if hash(f.Name) == id {
			file = f
			return storer.ErrStop
		}

		return nil
	})
	CheckError(err)

	return file
}

func (n *Notebook) Compact() {
	tree, err := n.Repository.Worktree()
	CheckError(err)

	wtroot := tree.Filesystem.Root()

	cmd := exec.Command("git", "-C", tree.Filesystem.Root(), "gc", "--aggressive", "--prune=now")
	err = cmd.Run()
	CheckError(err)

	n.Repository, err = LoadRepository(wtroot)
	CheckError(err)
}
